@@ -90,6 +90,37 @@ static char *do_peekstr(unsigned long lptr)
 	if (lptr < sizeof(long))
 		return NULL;
 
+#ifdef HAVE_PROCESS_VM_READV
+	struct iovec liov, riov;
+
+	/* We can't cross remote page boundaries when using this :( */
+	l = 0x1000;
+	riov.iov_base = (void *)lptr;
+	len = lptr % l;
+	if (!len)
+		len = l;
+	liov.iov_base = ret = xmalloc(len);
+	riov.iov_len = liov.iov_len = len;
+
+	while (1) {
+		if (process_vm_readv(trace_pid, &liov, 1, &riov, 1, 0) == -1) {
+			if (errno == ENOSYS)
+				break;
+			sb_ebort("ISE:do_peekstr: process_vm_readv() hates us: %s\n",
+				strerror(errno));
+		}
+
+		for (i = 0; i < liov.iov_len; ++i)
+			if (!((char *)liov.iov_base)[i])
+				return ret;
+		riov.iov_base += l;
+		riov.iov_len = liov.iov_len = l;
+		len += l;
+		ret = xrealloc(ret, len);
+		liov.iov_base = ret + len - l;
+	}
+#endif
+
 	l = 0;
 	len = 1024;
 	ret = xmalloc(len);
@@ -473,8 +504,9 @@ void trace_main(const char *filename, char *const argv[])
 		/* Not all kernel versions support this, so ignore return */
 		ptrace(PTRACE_SETOPTIONS, trace_pid, NULL, (void *)PTRACE_O_TRACESYSGOOD);
 #endif
+		sb_close_all_fds();
 		trace_loop();
-		return;
+		sb_ebort("ISE: child should have quit, as should we\n");
 	}
 
 	sb_debug("child setting up ...");
